"""GUI frontend for Posetta

The GUI frontend gives an alternative for those that are not comfortable using
the command line.

"""

from posetta.gui.posetta_gui import PosettaGui  # type: ignore  # noqa
